// $Id: Ellipsoids.h 900 2009-08-13 20:00:45Z larry $
//
// generated by Fast Light User Interface Designer (fluid) version 1.0104

#ifndef Ellipsoids_h
#define Ellipsoids_h
#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Check_Button.H>
#include <FL/Fl_Text_Display.H>
#include <FL/Fl_Text_Editor.H>
#include "Draw_Fl_Input.h"
#include <FL/Fl_Button.H>
#include <FL/Fl_Output.H>
#include "Flu_Combo_List.h"
void Include_Cutouts_cb (void);

void Edit_Ellipsoid_Close_cb (Fl_Button *, void *);

void Edit_Ellipsoid_Save_cb (Fl_Button *, int *);

class Ellipsoids
{
  public:
    Draw_Fl_Input * Probability;
    Draw_Fl_Input *Axis_Width;
    Flu_Combo_List *Axis_Color;
    Flu_Combo_List *Cutout_Color;
    Fl_Text_Buffer *ColorInputBuf;
      Ellipsoids ();
    Fl_Window *Ellips_Window;
    Fl_Check_Button *Show_Ellipsoids;
    Fl_Check_Button *Use_Cutouts;
    Fl_Text_Editor *Color_Input;
    Flu_Combo_List *Atom_Combo;
    Flu_Combo_List *Color_Combo;
    Fl_Button *New_Ellipsoid_Add;
    Fl_Button *New_Ellipsoid_Remove;
    Fl_Button *New_Ellipsoid_Convert;
    Fl_Output *Instr;
    Fl_Output *Instr1;
    Fl_Output *Instr2;

    void show (void);
};
#endif
