// $Id: EditView.h 900 2009-08-13 20:00:45Z larry $
//
// generated by Fast Light User Interface Designer (fluid) version 1.0104

#ifndef EditView_h
#define EditView_h
#include <stdlib.h>
#include <FL/Fl.H>
#include <FL/Fl_Check_Button.H>
#include <FL/Fl_Text_Display.H>
#include "Draw_Fl_Input.h"
#include <FL/Fl_Button.H>
#include <FL/Fl_Menu_.H>
#include "Flu_Combo_List.h"

void Clear_Last_Omit_cb (Fl_Button *, void *);

void Edit_Parmeters_Save_cb (Fl_Button *, int *);

void Edit_Parmeters_Close_cb (Fl_Button *, void *);

void Edit_Arrow_Close_cb (Fl_Button *, void *);

void Edit_Arrow_Save_cb (Fl_Button *, int *);

void Input_Help_cb (void);

class EditScreen
{
  public:
    Draw_Fl_Input * Magnification;
    Draw_Fl_Input *List;
    Draw_Fl_Input *Mol_Comp_Dist;
    Draw_Fl_Input *DepthCue;
    Flu_Combo_List *Cell_Edge_Color;
    Draw_Fl_Input *Cell_Edge_Width;
    Draw_Fl_Input *Poly_Edge_Color;
    Draw_Fl_Input *Poly_Edge_Width;
    Draw_Fl_Input *Poly_Limit;
    Draw_Fl_Input *Phong_Refl;
    Draw_Fl_Input *Phong_Size;
    Draw_Fl_Input *Ambient_Finish;
    Draw_Fl_Input *Diffuse_Finish;
    Draw_Fl_Input *Specular_Finish;
    Draw_Fl_Input *Finish_Roughness;
    Draw_Fl_Input *Label_Scale;
    Flu_Combo_List *Background_Color;
      EditScreen ();
    Fl_Window *editWindow;
    Fl_Check_Button *NoShadow;
    Fl_Check_Button *NoLabels;
    Fl_Check_Button *MolCompButton;
    Fl_Check_Button *ClearOmit;
    Fl_Button *ClearLastOmit;
  private:
      Fl_Button * Edit_Prm_Save;
    Fl_Button *Edit_Prm_Apply;
    Fl_Button *Edit_Prm_Close;
    Fl_Button *Edit_View_Help;
  public:
    void show (void);
};

class ArrowScreen
{
  public:
    Fl_Window * Arrow_View ();
    Fl_Window *arrowedit;
    Fl_Input *Px;
    Fl_Input *Py;
    Fl_Input *Pz;
    Fl_Input *Cz;
    Fl_Input *Cy;
    Fl_Input *Cx;
    Fl_Input *Aa;
    Fl_Input *Ab;
    Fl_Input *Ac;
    Fl_Input *Ba;
    Fl_Input *Bb;
    Fl_Input *Bc;
    Fl_Input *Ca;
    Fl_Input *Cb;
    Fl_Input *Cc;
};

#endif
